#!/usr/bin/env python3
# 
# pyalpm - Python 3 Bindings for libalpm
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
Lists optional dependencies of installed packages.

Options:
	-u : only list uninstalled optdepends
	-q : only print package names
"""

import re
import argparse

from pycman import config

re_depstring = re.compile("([^<=>]*)([<>]?=.*)?")

def enum_optdepends(h):
	for p in h.get_localdb().pkgcache:
		for o in p.optdepends:
			oname, _colon, odesc = o.partition(':')
			yield (p.name, oname.strip(), odesc.strip())

def enum_depends(h):
	for p in h.get_localdb().pkgcache:
		for d in p.depends:
			m = re_depstring.match(d)
			yield (p.name, m.group(1))

def main(opts=None):
	if opts is None:
		opts = parse_options()

	h = config.init_with_config("/etc/pacman.conf")
	optdeps = {}
	installed = set(p.name for p in h.get_localdb().pkgcache)
	needed = set(q for (p, q) in enum_depends(h))

	for suggester, optdep, reason in enum_optdepends(h):
		optdeps.setdefault(optdep, {})[suggester] = reason

	for optdep in sorted(optdeps):
		if optdep in installed and opts.u:
			continue
		if opts.orphan and not (optdep in installed and optdep not in needed):
			continue
		suggesters = optdeps[optdep]
		if opts.q:
			print(optdep)
		else:
			print(optdep, 'is suggested by')
			for name, reason in sorted(suggesters.items()):
				print('    ', name, ':', reason)

def parse_options():
	p = argparse.ArgumentParser()
	p.add_argument('-u', action = 'store_true', default = False,
			help='only list uninstalled optdepends')
	p.add_argument('--orphan', action = 'store_true', default = False,
			help='list installed optdepends that are not hard deps of any package')
	p.add_argument('-q', action = 'store_true', default = False,
			help='only print package names')
	return p.parse_args()

if __name__ == "__main__":
	main(parse_options())

# vim: set ts=4 sw=4 tw=0 noet:



#!/usr/bin/env python3
#
# pycman-sync - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
A Python implementation of pacman -S

This script displays information about packages available in repositories,
and is also used to install/upgrade/remove them.
"""

import sys

import pyalpm
from pycman import config
from pycman import pkginfo
from pycman import transaction

handle = None

def do_clean(options):
	raise NotImplementedError

def do_refresh(options):
	"Sync databases like pacman -Sy"
	force = (options.refresh > 1)
	for db in handle.get_syncdbs():
		t = transaction.init_from_options(handle, options)
		db.update(force)
		t.release()
	return 0

def do_sysupgrade(options):
	"Upgrade a system like pacman -Su"
	downgrade = (options.sysupgrade > 1)
	t = transaction.init_from_options(handle, options)
	t.sysupgrade(downgrade)
	if len(t.to_add) + len(t.to_remove) == 0:
		print("nothing to do")
		t.release()
		return 0
	else:
		ok = transaction.finalize(t)
		return (0 if ok else 1)

def do_install(pkgs, options):
	"Install a list of packages like pacman -S"
	repos = dict((db.name, db) for db in handle.get_syncdbs())
	if len(pkgs) == 0:
		print("error: no targets specified")
		return 1

	targets = []
	for name in pkgs:
		ok, pkg = find_sync_package(name, repos)
		if not ok:
			print('error:', pkg)
			return 1
		else:
			targets.append(pkg)
	t = transaction.init_from_options(handle, options)
	[t.add_pkg(pkg) for pkg in targets]
	ok = transaction.finalize(t)
	return (0 if ok else 1)

def find_sync_package(pkgname, syncdbs):
	"Finds a package name of the form 'repo/pkgname' or 'pkgname' in a list of DBs"
	if '/' in pkgname:
		repo, pkgname = pkgname.split('/', 1)
		db = syncdbs.get(repo)
		if db is None:
			return False, "repository '%s' does not exist" % repo
		pkg = db.get_pkg(pkgname)
		if pkg is None:
			return False, "package '%s' was not found in repository '%s'" % (pkgname, repo)
		return True, pkg
	else:
		for db in syncdbs.values():
			pkg = db.get_pkg(pkgname)
			if pkg is not None:
				return True, pkg
		return False, "package '%s' was not found" % pkgname

# Query actions

def show_groups(args):
	"Show groups like pacman -Sg"
	for repo in handle.get_syncdbs():
		if len(args.args) == 0:
			# list all available groups
			[print(name) for name, pkgs in repo.grpcache]
		else:
			# only print chosen groups
			for group in args.args:
				grp = repo.read_grp(group)
				if grp is None:
					continue
				else:
					name, pkgs = grp
				if args.quiet:
					[print(pkg.name) for pkg in pkgs]
				else:
					[print(name, pkg.name) for pkg in pkgs]
	return 0

def show_repo(args):
	"Show repository's list of packages like pacman -Sl"
	repos = handle.get_syncdbs()
	if len(args.args) > 0:
		repo_dict = dict((repo.name, repo) for repo in repos)
		try:
			repos = [repo_dict[name] for name in args.args]
		except KeyError as err:
			print("error: repository '%s' was not found" % err.args)
			return 1

	for repo in repos:
		if args.quiet:
			[print(pkg.name) for pkg in repo.pkgcache]
		else:
			[print(repo.name, pkg.name, pkg.version) for pkg in repo.pkgcache]
	return 0

def show_packages(args):
	"Show information about packages like pacman -Si"
	retcode = 0
	if len(args.args) == 0:
		for repo in handle.get_syncdbs():
			for pkg in repo.pkgcache:
				pkginfo.display_pkginfo(pkg, level=args.info, style='sync')
	else:
		repos = dict((db.name, db) for db in handle.get_syncdbs())
		for pkgname in args.args:
			ok, value = find_sync_package(pkgname, repos)
			if ok:
				pkginfo.display_pkginfo(value, level=args.info, style='sync')
			else:
				retcode = 1
				print("error:", value)
	return retcode

def show_search(patterns, options):
	results = []
	for db in handle.get_syncdbs():
		results += db.search(*patterns)
	if len(results) == 0:
		return 1
	for pkg in results:
		if options.quiet:
			print(pkg.name)
		else:
			print("%s/%s %s" % (pkg.db.name, pkg.name, pkg.version))
			print("    " + pkg.desc)
	return 0

def parse_options(rawargs):
	parser = config.make_parser(prog='pycman-sync')
	# Misc actions
	group0 = parser.add_argument_group("Actions (default is installing specified packages)")
	group0.add_argument("-c", "--clean",
			action='count', default=0,
			help='remove old packages from cache directory (-cc for all)')
	group0.add_argument("-u", "--sysupgrade",
			action='count', default=0,
			help='upgrade installed packages (-uu allows downgrade)')
	group0.add_argument("-y", "--refresh",
			action='count', default=0,
			help='download fresh package databases from the server')
	# Installation options
	grp_install = parser.add_argument_group("Install options")
	grp_install.add_argument('-d', '--nodeps',
			action='store_true', default=False,
			help='skip dependency checks')
	grp_install.add_argument('-f', '--force',
			action='store_true', default=False,
			help='force install, overwrite conflicting files')
	grp_install.add_argument('-k', '--dbonly',
			action='store_true', default=False,
			help='only modify database entries, not package files')
	grp_install.add_argument('-w', '--downloadonly',
			action='store_true', default=False,
			help='download packages but do not install/upgrade anything')
	grp_install.add_argument('--asdeps', dest='mode',
			action="store_const",
			const=pyalpm.PKG_REASON_DEPEND)
	grp_install.add_argument('--asexplicit', dest='mode',
			action="store_const",
			const=pyalpm.PKG_REASON_EXPLICIT)
	# Options to query sync databases
	group1 = parser.add_argument_group("Query actions")
	group1.add_argument('-g', '--groups', action='store_true', default=False,
			help='view list of groups, or all members of a package group')
	group1.add_argument('-i', '--info',
			action='count', dest='info', default=0,
			help='view package information')
	group1.add_argument('-l', '--list', action='store_true', default=False,
			help='list the contents of repositories')
	group1.add_argument('-s', '--search', action='store_true', default=False,
			help='search remote repositories for matching strings')
	group1.add_argument('-q', '--quiet',
			action='store_true', dest='quiet', default=False,
			help='show less information for query and search')
	group1.add_argument('args', metavar='arg', nargs='*',
			help='arguments (group names for -g, repo names for -l, '
			'package names for -i)')
	return parser.parse_args(rawargs)

def main(rawargs=None):
	if rawargs is None:
		rawargs = sys.argv[1:]

	global handle
	args = parse_options(rawargs)
	handle = config.init_with_config_and_options(args)

	if args.verbose:
		print("sync " + " ".join(rawargs), file=sys.stderr)

	# Refresh databases if necessary
	if args.refresh > 0:
		ret = do_refresh(args)
		if ret != 0:
			return ret

	# If a query action is set
	if args.search:
		return show_search(args.args, args)
	elif args.groups:
		return show_groups(args)
	elif args.info:
		return show_packages(args)
	elif args.list:
		return show_repo(args)
	# If a cleanup is required
	elif args.clean > 0:
		return do_clean(args)
	# If a sysupgrade is required
	elif args.sysupgrade > 0:
		return do_sysupgrade(args)
	# If only a refresh was requested
	elif len(args.args) == 0 and args.refresh > 0:
		return 0
	# Otherwise it's a normal install
	else:
		return do_install(args.args, args)

if __name__ == "__main__":
	ret = main(sys.argv[1:])
	sys.exit(ret)

# vim: set ts=4 sw=4 tw=0 noet:



#!/usr/bin/env python3
#
# pycman - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
pycman configuration handling

This module handles pacman.conf files as well as pycman options that
are common to all action modes.
"""

import os
import glob
import sys
import argparse
import collections
import warnings

import pyalpm

class InvalidSyntax(Warning):
	def __init__(self, filename, problem, arg):
		self.filename = filename
		self.problem = problem
		self.arg = arg

	def __str__(self):
		return "unable to parse %s, %s: %r" % (self.filename, self.problem, self.arg)

# Options that may occur several times in a section. Their values should be
# accumulated in a list.
LIST_OPTIONS = (
	'CacheDir',
	'HookDir',
	'HoldPkg',
	'SyncFirst',
	'IgnoreGroup',
	'IgnorePkg',
	'NoExtract',
	'NoUpgrade',
	'Server'
)

SINGLE_OPTIONS = (
	'RootDir',
	'DBPath',
	'GPGDir',
	'LogFile',
	'Architecture',
	'XferCommand',
	'CacheServer',
	'CleanMethod',
	'SigLevel',
	'LocalFileSigLevel',
	'RemoteFileSigLevel',
	'ParallelDownloads',
	'DownloadUser',
)

BOOLEAN_OPTIONS = (
	'UseSyslog',
	'ShowSize',
	'CheckSpace',
	'VerbosePkgLists',
	'ILoveCandy',
	'Color',
	'DisableDownloadTimeout',
	'NoProgressBar',
	'DisableSandbox',
	'DisableSandboxFilesystem',
	'DisableSandboxSyscalls',
)


class PacmanConfEnumeratorSession():

	def __init__(self, path):
		self.path = path
		self.file_descriptors = []

	def _enumerator(self):
		filestack = []
		current_section = None
		new_fd = open(self.path, 'rb')
		self.file_descriptors.append(new_fd)
		filestack.append(new_fd)
		while len(filestack) > 0:
			f = filestack[-1]
			line = f.readline()
			line = line.decode('utf-8')
			if len(line) == 0:
				# end of file
				filestack.pop()
				continue

			line = line.strip()
			if len(line) == 0:
				continue
			if line[0] == '#':
				continue
			if line[0] == '[' and line[-1] == ']':
				current_section = line[1:-1]
				continue
			# read key, value
			key, equal, value = [x.strip() for x in line.partition('=')]

			if current_section is None and key != 'Include':
				raise InvalidSyntax(f.name, 'statement outside of a section', line)

			# include files
			if equal == '=' and key == 'Include':
				new_fds = [open(f, 'rb') for f in glob.glob(value)]
				filestack.extend(new_fds)
				self.file_descriptors.extend(new_fds)
				continue
			if current_section != 'options':
				# repos only have the Server, SigLevel, Usage options
				if key in ('Server', 'SigLevel', 'Usage', 'CacheServer') and equal == '=':
					yield (current_section, key, value)
				else:
					raise InvalidSyntax(f.name, 'invalid key for repository configuration', line)
				continue
			if equal == '=':
				if key in LIST_OPTIONS:
					for val in value.split():
						yield (current_section, key, val)
				elif key in SINGLE_OPTIONS:
					yield (current_section, key, value)
				else:
					warnings.warn(InvalidSyntax(f.name, 'unrecognized option', key))
			else:
				if key in BOOLEAN_OPTIONS:
					yield (current_section, key, True)
				else:
					warnings.warn(InvalidSyntax(f.name, 'unrecognized option', key))

	def __enter__(self):
		return self._enumerator

	def __exit__(self, *exc_details):
		for fd in self.file_descriptors:
			fd.close()


_logmask = pyalpm.LOG_ERROR | pyalpm.LOG_WARNING

def cb_log(level, line):
	if not (level & _logmask):
		return
	if level & pyalpm.LOG_ERROR:
		line = "ERROR: " + line
	elif level & pyalpm.LOG_WARNING:
		line = "WARNING: " + line
	elif level & pyalpm.LOG_DEBUG:
		line = "DEBUG: " + line
	elif level & pyalpm.LOG_FUNCTION:
		line = "FUNC: " + line
	sys.stderr.write(line)

class PacmanConfig(object):
	def __init__(self, conf=None, options=None):
		self.options = {}
		self.repos = collections.OrderedDict()
		self.options["RootDir"] = "/"
		self.options["DBPath"] = "/var/lib/pacman"
		self.options["GPGDir"] = "/etc/pacman.d/gnupg/"
		self.options["LogFile"] = "/var/log/pacman.log"
		self.options["Architecture"] = os.uname()[-1]
		if conf is not None:
			self.load_from_file(conf)
		if options is not None:
			self.load_from_options(options)

	def load_from_file(self, filename):
		with PacmanConfEnumeratorSession(filename) as pacman_conf_enumerator:
			for section, key, value in pacman_conf_enumerator():
				if section == 'options':
					if key == 'Architecture' and value == 'auto':
						continue
					if key in LIST_OPTIONS:
						self.options.setdefault(key, []).append(value)
					else:
						self.options[key] = value
				else:
					servers = self.repos.setdefault(section, [])
					if key == 'Server':
						servers.append(value)
		if "CacheDir" not in self.options:
			self.options["CacheDir"] = ["/var/cache/pacman/pkg"]

	def load_from_options(self, options):
		global _logmask
		if options.root is not None:
			self.options["RootDir"] = options.root
		if options.dbpath is not None:
			self.options["DBPath"] = options.dbpath
		if options.gpgdir is not None:
			self.options["GPGDir"] = options.gpgdir
		if options.arch is not None:
			self.options["Architecture"] = options.arch
		if options.logfile is not None:
			self.options["LogFile"] = options.logfile
		if options.cachedir is not None:
			self.options["CacheDir"] = [options.cachedir]
		if options.debug:
			_logmask = 0xffff

	def apply(self, h):
		# File paths
		h.logfile = self.options["LogFile"]
		h.gpgdir = self.options["GPGDir"]
		# Lists
		arch = self.options["Architecture"]
		if isinstance(arch, str):
			arch = [arch]
		h.arch = arch
		h.cachedirs = self.options["CacheDir"]
		if "NoUpgrade" in self.options:
			h.noupgrades = self.options["NoUpgrade"]
		if "NoExtract" in self.options:
			h.noextracts = self.options["NoExtract"]
		if "IgnorePkg" in self.options:
			h.ignorepkgs = self.options["IgnorePkg"]
		if "IgnoreGroup" in self.options:
			h.ignoregrps = self.options["IgnoreGroup"]

		h.logcb = cb_log

		# set sync databases
		for repo, servers in self.repos.items():
			db = h.register_syncdb(repo, 0)
			db_servers = []
			for rawurl in servers:
				url = rawurl.replace("$repo", repo)
				url = url.replace("$arch", self.options["Architecture"])
				db_servers.append(url)
			db.servers = db_servers

	def initialize_alpm(self):
		h = pyalpm.Handle(self.options["RootDir"], self.options["DBPath"])
		self.apply(h)
		return h

	def __str__(self):
		return("PacmanConfig(options=%s, repos=%s)" % (str(self.options), str(self.repos)))

def make_parser(*args, **kwargs):
	parser = argparse.ArgumentParser(*args, **kwargs)
	common = parser.add_argument_group('Common options')
	common.add_argument('-b', '--dbpath', metavar='<path>',
			action='store', dest='dbpath', type=str,
			help='set an alternate database location')
	common.add_argument('-r', '--root', metavar='<path>',
			action='store', dest='root', type=str,
			help='set an alternate installation root')
	common.add_argument('-v', '--verbose',
			action='store_true', dest='verbose', default=False,
			help='be verbose')
	common.add_argument('--arch', metavar='<arch>',
			action='store', dest='arch', type=str,
			help='set an alternate architecture')
	common.add_argument('--config', metavar='<file>',
			action='store', dest='config', type=str,
			help='set an alternate configuration file')
	common.add_argument('--debug',
			action='store_true', dest='debug',
			help='display debug messages')
	common.add_argument('--logfile', metavar='<file>',
			action='store', dest='logfile', type=str,
			help='set an alternate log file')
	common.add_argument('--gpgdir', metavar='<dir>',
			action='store', dest='gpgdir', type=str,
			help='set an alternate log file')
	common.add_argument('--cachedir', metavar='<dir>',
			action='store', dest='cachedir', type=str,
			help='set an alternate cche location')
	return parser

def init_with_config(configpath):
	"Reads configuration from given path and apply it to libalpm"
	config = PacmanConfig(conf=configpath)
	return config.initialize_alpm()

def init_with_config_and_options(options):
	"Reads configuration from file and commandline options, and apply it to libalpm"
	# read config file
	if options.config is not None:
		config_file = options.config
	else:
		config_file = "/etc/pacman.conf"

	conf = PacmanConfig(conf=config_file, options=options)
	return conf.initialize_alpm()

# vim: set ts=4 sw=4 tw=0 noet:



#!/usr/bin/env python3
#
# pycman - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
pycman transaction management

This module defines convenient wrappers around pyalpm functions
to initialize transactions according to options
"""

import math
import sys
import traceback
import pyalpm

# Callbacks
def cb_event(*args):
	print("event", args)

def cb_conv(*args):
	print("conversation", args)

last_target = None
last_percent = 100
last_i = -1
def cb_progress(target, percent, n, i):
	"Display progress percentage for target i/n"
	global last_target, last_percent, last_i
	if len(target) == 0:
		# abstract progress
		if percent < last_percent or i < last_i:
			sys.stdout.write("progress (%d targets)" % n)
			last_i = 0
		sys.stdout.write((i - last_i) * '.')
		sys.stdout.flush()
		last_i = i
	else:
		# progress for some target (write 25 dots for 100%)
		if target != last_target or percent < last_percent:
			last_target = target
			last_percent = 0
			sys.stdout.write("progress for %s (%d/%d)" % (target, i, n))
		old_dots = last_percent // 4
		new_dots = percent // 4
		sys.stdout.write((new_dots - old_dots) * '.')
		sys.stdout.flush()

	# final newline
	if percent == 100 and last_percent < 100:
		sys.stdout.write('\n')
		sys.stdout.flush()
	last_percent = percent

_last_dl_filename = None
_last_dl_progress = None
_last_dl_total = None
def cb_dl(filename, tx, total):
	global _last_dl_filename, _last_dl_progress, _last_dl_total
	# check if a new file is coming
	if filename != _last_dl_filename or _last_dl_total != total:
		_last_dl_filename = filename
		_last_dl_total = total
		_last_dl_progress = 0
		sys.stdout.write("\ndownload %s: %d/%d" % (filename, tx, total))
		sys.stdout.flush()
	# compute a progress indicator
	if _last_dl_total > 0:
		progress = (tx * 25) // _last_dl_total
	else:
		# if total is unknown, use log(kBytes)²/2
		progress = int(math.log(1 + tx / 1024) ** 2 / 2)
	if progress > _last_dl_progress:
		_last_dl_progress = progress
		sys.stdout.write("\rdownload %s: %s %d/%d" % (filename, '.' * progress, tx, total))
		sys.stdout.flush()

def init_from_options(handle, options):
	"Transaction initialization"
	handle.dlcb = cb_dl
	handle.eventcb = cb_event
	handle.questioncb = cb_conv
	handle.progresscb = cb_progress
	t = handle.init_transaction(
			cascade=getattr(options, "cascade", False),
			nodeps=getattr(options, "nodeps", False),
			force=getattr(options, 'force', False),
			dbonly=getattr(options, 'dbonly', False),
			downloadonly=getattr(options, 'downloadonly', False),
			nosave=getattr(options, 'nosave', False),
			recurse=(getattr(options, 'recursive', 0) > 0),
			recurseall=(getattr(options, 'recursive', 0) > 1),
			unneeded=getattr(options, 'unneeded', False),
			alldeps=(getattr(options, 'mode', None) == pyalpm.PKG_REASON_DEPEND),
			allexplicit=(getattr(options, 'mode', None) == pyalpm.PKG_REASON_EXPLICIT))
	return t

def finalize(t):
	"Commit a transaction"
	try:
		t.prepare()
		t.commit()
	except pyalpm.error:
		traceback.print_exc()
		t.release()
		return False
	t.release()
	return True

# vim: set ts=4 sw=4 tw=0 noet:



#!/usr/bin/env python3
#
# pycman-remove - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
A Python implementation of pacman -R

This script uninstalls packages. Various options control
the effect on dependencies of/on given targets.
"""

import sys
from pycman import config
from pycman import transaction

handle = None

def remove(pkgs, options):
	# prepare target list
	db = handle.get_localdb()
	targets = []
	for name in pkgs:
		pkg = db.get_pkg(name)
		if pkg is None:
			print("error: '%s': target not found" % name)
			return 1
		targets.append(pkg)

	t = transaction.init_from_options(handle, options)

	for pkg in targets:
		t.remove_pkg(pkg)

	ok = transaction.finalize(t)
	return (0 if ok else 1)

def main(rawargs=None):
	if rawargs is None:
		rawargs = sys.argv[1:]

	global handle
	parser = config.make_parser()
	group = parser.add_argument_group("Remove options")
	group.add_argument('-c', '--cascade',
			action='store_true', default=False,
			help='remove packages and all packages that depend on them')
	group.add_argument('-d', '--nodeps',
			action='store_true', default=False,
			help='skip dependency checks')
	group.add_argument('-k', '--dbonly',
			action='store_true', default=False,
			help='only modify database entries, not package files')
	group.add_argument('-n', '--nosave',
			action='store_true', default=False,
			help='remove configuration files as well')
	group.add_argument('-s', '--recursive',
			action='store_true', default=False,
			help="remove dependencies also (that won't break packages)")
	group.add_argument('-u', '--unneeded',
			action='store_true', default=False,
			help="remove unneeded packages (that won't break packages)")
	group.add_argument('pkgs', metavar='pkg', nargs='*',
			help="a list of packages, e.g. libreoffice, openjdk6")

	args = parser.parse_args(rawargs)
	handle = config.init_with_config_and_options(args)

	if args.verbose:
		print("remove " + " ".join(rawargs), file=sys.stderr)

	if len(args.pkgs) == 0:
		print('error: no targets specified')
		return 1

	return remove(args.pkgs, args)

if __name__ == "__main__":
	ret = main(sys.argv[1:])
	sys.exit(ret)

# vim: set ts=4 sw=4 tw=0 noet:



#!/usr/bin/env python3
#
# pycman-deptest - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
A Python implementation of pacman -T

This script checks whether specified dependencies are satisfied
and prints out a list of those which are missing.
"""

import sys
import pyalpm
from pycman import config

handle = None

def deptest(deps):
	db = handle.get_localdb()
	missing = [dep for dep in deps if pyalpm.find_satisfier(db.pkgcache, dep) is None]
	return missing

def main(rawargs=None):
	if rawargs is None:
		rawargs = sys.argv[1:]

	global handle
	parser = config.make_parser()
	parser.add_argument('deps', metavar='dep', nargs='*',
			help="a dependency string, e.g. 'pacman>=3.4.0'")
	args = parser.parse_args(rawargs)
	handle = config.init_with_config_and_options(args)

	if args.verbose:
		print("deptest " + " ".join(rawargs), file=sys.stderr)
	missing = deptest(args.deps)

	if len(missing) == 0:
		return 0
	else:
		[print(dep) for dep in missing]
		return 127

if __name__ == "__main__":
	ret = main(sys.argv[1:])
	sys.exit(ret)

# vim: set ts=4 sw=4 tw=0 noet:



#!/usr/bin/env python3
#
# pycman-query - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
A Python implementation of pacman -Q

This script displays information about installed packages.
"""

import os
import sys

import pyalpm
from pycman import config
from pycman import pkginfo

handle = None

def filter_pkglist(pkglist, options):
	result = []
	if options.foreign:
		syncpkgs = set()
		for db in handle.get_syncdbs():
			syncpkgs |= set(p.name for p in db.pkgcache)
	for pkg in pkglist:
		if options.deps and pkg.reason == pyalpm.PKG_REASON_EXPLICIT:
			continue
		if options.explicit and pkg.reason == pyalpm.PKG_REASON_DEPEND:
			continue
		if options.unrequired and len(pkg.compute_requiredby()) > 0:
			continue
		if options.foreign and pkg.name in syncpkgs:
			continue
		if options.upgrades and pyalpm.sync_newversion(pkg, handle.get_syncdbs()) is None:
			continue
		result.append(pkg)
	return result

def display_pkg(pkg, options):
	displaystyle = 'file' if options.package else 'local'
	if options.info > 0:
		pkginfo.display_pkginfo(pkg, level=options.info, style=displaystyle)
	elif not options.listfiles:
		if options.quiet:
			print(pkg.name)
		else:
			print(pkg.name, pkg.version)

	if options.listfiles:
		if options.quiet:
			[print('/' + path) for path, size, mode in pkg.files]
		else:
			[print(pkg.name, '/' + path) for path, size, mode in pkg.files]

def find_file(filenames, options):
	"lookup for files in local packages"
	ret = 0
	if len(filenames) == 0:
		print("error: no targets specified")
		ret = 1

	localpkgs = handle.get_localdb().pkgcache

	for name in filenames:
		lookupname = None
		if not os.path.isabs(name):
			# lookup in PATH
			for dirname in os.getenv('PATH').split(':'):
				if os.path.lexists(os.path.join(dirname, name)):
					name = os.path.join(dirname, name)
					lookupname = name
			if lookupname is None:
				print("error: failed to find '%s' in PATH: No such file or directory" % name)
				ret = 1
				continue
		else:
			if not os.path.lexists(name):
				print("error: failed to read file '%s': No such file or directory" % name)
				ret = 1
				continue
			lookupname = name
		lookupname = os.path.normpath(lookupname)
		lookupname = lookupname.lstrip('/')
		found = False
		for pkg in localpkgs:
			files = pkg.files

			if any(f for f in files if lookupname == f[0]):
				if options.quiet:
					print(pkg.name)
				else:
					print(pkg.name, "is owned by", pkg.name, pkg.version)
				found = True
				break

		if not found:
			print('error: no package owns', name)
			ret = 1

	return ret

def find_search(patterns, options):
	db = handle.get_localdb()
	results = db.search(*patterns)
	if len(results) == 0:
		return 1
	for pkg in results:
		if options.quiet:
			print(pkg.name)
		else:
			print("%s/%s %s" % (pkg.db.name, pkg.name, pkg.version))
			print("    " + pkg.desc)
	return 0

def main(rawargs=None):
	if rawargs is None:
		rawargs = sys.argv[1:]

	global handle
	parser = config.make_parser(prog='pycman-query')
	group = parser.add_argument_group("Query options")
	group.add_argument('-d', '--deps',
			action='store_true', default=False,
			help='list packages installed as dependencies [filter]')
	group.add_argument('-e', '--explicit',
			action='store_true', default=False,
			help='list packages explicitly installed [filter]')
	group.add_argument('-i', '--info',
			action='count', dest='info', default=0,
			help='view package information')
	group.add_argument('-l', '--list',
			action='store_true', dest='listfiles', default=False,
			help='list the contents of the queried package')
	group.add_argument('-m', '--foreign',
			action='store_true', default=False,
			help='list installed packages not found in sync db(s) [filter]')
	group.add_argument('-o', '--owns',
			action='store_true', default=False,
			help='query the package that owns <file>')
	group.add_argument('-p', '--package',
			action='store_true', default=False,
			help='query a package file instead of the database')
	group.add_argument('-q', '--quiet',
			action='store_true', dest='quiet', default=False,
			help='show less information for query and search')
	group.add_argument('-s', '--search', action='store_true', default=False,
			help='search locally-installed packages for matching strings')
	group.add_argument('-t', '--unrequired',
			action='store_true', default=False,
			help="list packages not required by any package [filter]")
	group.add_argument('-u', '--upgrades',
			action='store_true', default=False,
			help="list outdated packages [filter]")
	group.add_argument('pkgnames', metavar='pkg', nargs='*',
			help='packages to show (show all packages if no arguments) '
			'(when used with -o: a filename, '
			'when used with -p: the path to a package file)')

	args = parser.parse_args(rawargs)
	handle = config.init_with_config_and_options(args)

	if args.verbose:
		print("query " + " ".join(rawargs), file=sys.stderr)

	db = handle.get_localdb()
	retcode = 0

	# actions other than listing packages
	if args.owns:
		return find_file(args.pkgnames, args)
	if args.search:
		return find_search(args.pkgnames, args)

	pkglist = []
	if len(args.pkgnames) > 0:
		# a list of package names was specified
		for pkgname in args.pkgnames:
			if args.package:
				pkg = handle.load_pkg(pkgname)
			else:
				pkg = db.get_pkg(pkgname)
			if pkg is None:
				print('error: package "%s" not found' % pkgname)
				retcode = 1
			else:
				pkglist.append(pkg)
	else:
		# no package was specified, display all
		pkglist = db.pkgcache
	# determine the list of package to actually display
	pkglist = filter_pkglist(pkglist, args)
	for pkg in pkglist:
		display_pkg(pkg, args)

	return retcode

if __name__ == "__main__":
	ret = main(sys.argv[1:])
	sys.exit(ret)

# vim: set ts=4 sw=4 tw=0 noet:



#!/usr/bin/env python3
#
# pycman.pkginfo - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
Package information formatting

This module defines utility function to format package information
for terminal output.
"""

import sys
import time
import textwrap

import struct
import fcntl
import termios

import pyalpm

ATTRNAME_FORMAT = '%-14s : '
ATTR_INDENT = 17 * ' '

def get_term_size():
	if sys.stdout.isatty():
		empty = struct.pack("HHHH", 0, 0, 0, 0)
		_, width, _, _ = struct.unpack("HHHH", fcntl.ioctl(1, termios.TIOCGWINSZ, empty))
		return width
	else:
		return 80

def format_attr(attrname, value, format_str=None):
	if isinstance(value, list):
		if len(value) == 0:
			valuestring = 'None'
		else:
			valuestring = '  '.join(str(v) for v in value)
	else:
		if format_str == "time":
			valuestring = time.strftime("%a %d %b %Y %X %Z", time.localtime(value))
		else:
			valuestring = str(value)
	return textwrap.fill(valuestring, width=get_term_size(),
			initial_indent=ATTRNAME_FORMAT % attrname,
			subsequent_indent=ATTR_INDENT,
			break_on_hyphens=False,
			break_long_words=False)

def format_attr_oneperline(attrname, value):
	if len(value) == 0:
		value = ['None']
	s = ATTRNAME_FORMAT % attrname
	s += ('\n' + ATTR_INDENT).join(value)
	return s

def display_pkginfo(pkg, level=1, style='local'):
	"""
	Displays pretty-printed package information.

	Args :
	  pkg -- the package to display
	  level -- the level of detail (1 or 2)
	  style -- 'local' or 'sync'
	"""

	if style not in ['local', 'sync', 'file']:
		raise ValueError('Invalid style for package info formatting')

	if style == 'sync':
		print(format_attr('Repository', pkg.db.name))
	print(format_attr('Name', pkg.name))
	print(format_attr('Base', pkg.base))
	print(format_attr('Version', pkg.version))
	print(format_attr('URL', pkg.url))
	print(format_attr('Licenses', pkg.licenses))
	print(format_attr('Groups', pkg.groups))
	print(format_attr('Provides', pkg.provides))
	print(format_attr('Depends On', pkg.depends))
	print(format_attr_oneperline('Optional Deps', pkg.optdepends))
	if style == 'local' or level == 2:
		print(format_attr('Required By', pkg.compute_requiredby()))
		print(format_attr('Optional For', pkg.compute_optionalfor()))
	print(format_attr('Conflicts With', pkg.conflicts))
	print(format_attr('Replaces', pkg.replaces))
	if style == 'sync':
		print(format_attr('Download Size', '%.2f K' % (pkg.size / 1024)))
	if style == 'file':
		print(format_attr('Compressed Size', '%.2f K' % (pkg.size / 1024)))
	print(format_attr('Installed Size', '%.2f K' % (pkg.isize / 1024)))
	print(format_attr('Packager', pkg.packager))
	print(format_attr('Architecture', pkg.arch))
	print(format_attr('Build Date', pkg.builddate, format_str='time'))

	if style == 'local':
		# local installation information
		print(format_attr('Install Date', pkg.installdate, format_str='time'))
		if pkg.reason == pyalpm.PKG_REASON_EXPLICIT:
			reason = 'Explicitly installed'
		elif pkg.reason == pyalpm.PKG_REASON_DEPEND:
			reason = 'Installed as a dependency for another package'
		else:
			reason = 'N/A'
		print(format_attr('Install Reason', reason))
	if style != 'sync':
		print(format_attr('Install Script', 'Yes' if pkg.has_scriptlet else 'No'))
	if style == 'sync':
		print(format_attr('MD5 Sum', pkg.md5sum))
		print(format_attr('SHA256 Sum', pkg.sha256sum))
		print(format_attr('Signatures', 'Yes' if pkg.base64_sig else 'No'))

	print(format_attr('Description', pkg.desc))

	if level >= 2 and style == 'local':
		# print backup information
		print('Backup files:')
		if len(pkg.backup) == 0:
			print('(none)')
		else:
			print('\n'.join(["%s %s" % (md5, file) for (file, md5) in pkg.backup]))
	print('')

# vim: set ts=4 sw=4 noet:



#!/usr/bin/env python3
#
# pycman - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from . import action_database
from . import action_deptest
from . import action_query
from . import action_remove
from . import action_sync
from . import action_upgrade
from . import action_version

ACTIONS = {
	'db': action_database.main,
	'deptest': action_deptest.main,
	'query': action_query.main,
	'remove': action_remove.main,
	'sync': action_sync.main,
	'upgrade': action_upgrade.main,
	'version': action_version.main
}

def run_action_with_args(action, args):
	if action not in ACTIONS:
		print("Invalid action specified (%s are supported)" % ', '.join(ACTIONS.keys()))
		return 1
	else:
		callback = ACTIONS[action]
		return callback(args)

# vim: set ts=4 sw=4 tw=0 noet:




#!/usr/bin/env python3
#
# pycman-version - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
A Python implementation of pacman -V

This script prints version information about pycman and pyalpm.
"""

import pyalpm

VERSION_STRING = """
                       Pycman running on pyalpm v%s - libalpm v%s
 .--.
/ _.-' .-.  .-.  .-.   libalpm:
\\  '-. '-'  '-'  '-'     Copyright (C) 2006-2011 Pacman Development Team
 '--'                    Copyright (C) 2002-2006 Judd Vinet
                       pyalpm:
                         Copyright (C) 2011 Rémy Oudompheng

                       This program may be freely redistributed under
                       the terms of the GNU General Public License.
"""

def main(args=None):
	print(VERSION_STRING % (pyalpm.version(), pyalpm.alpmversion()))
	return 2

if __name__ == "__main__":
	import sys
	ret = main(sys.argv[1:])
	sys.exit(ret)

# vim: set ts=4 sw=4 tw=0 noet:



#!/usr/bin/env python3
#
# pycman-upgrade - A Python implementation of Pacman
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
A Python implementation of pacman -U

This script installs packages from tarballs. Various options control
the effect of the transaction.
"""

import sys
import pyalpm
from pycman import config
from pycman import transaction

handle = None

def upgrade(pkgs, options):
	# prepare target list
	targets = []
	for name in pkgs:
		pkg = handle.load_pkg(name)
		targets.append(pkg)

	t = transaction.init_from_options(handle, options)

	for pkg in targets:
		t.add_pkg(pkg)

	ok = transaction.finalize(t)
	return (0 if ok else 1)

def main(rawargs=None):
	if rawargs is None:
		rawargs = sys.argv[1:]

	global handle
	parser = config.make_parser()
	group = parser.add_argument_group("upgrade options")
	group.add_argument('-d', '--nodeps',
			action='store_true', default=False,
			help='skip dependency checks')
	group.add_argument('-f', '--force',
			action='store_true', default=False,
			help='force install, overwrite conflicting files')
	group.add_argument('-k', '--dbonly',
			action='store_true', default=False,
			help='only modify database entries, not package files')
	group.add_argument('--asdeps', dest='mode',
			action="store_const",
			const=pyalpm.PKG_REASON_DEPEND)
	group.add_argument('--asexplicit', dest='mode',
			action="store_const",
			const=pyalpm.PKG_REASON_EXPLICIT)
	group.add_argument('pkgs', metavar='pkg', nargs='*',
			help="a list of package URLs, e.g. package-1.0-1-x86_64.tar.xz")

	args = parser.parse_args(rawargs)
	handle = config.init_with_config_and_options(args)

	if args.verbose:
		print("upgrade " + " ".join(rawargs), file=sys.stderr)

	return upgrade(args.pkgs, args)

if __name__ == "__main__":
	ret = main(sys.argv[1:])
	sys.exit(ret)

# vim: set ts=4 sw=4 tw=0 noet:
